import unittest
from unittest.mock import patch

from soil.aio.dictionary import dictionary


class TestDictionary(unittest.IsolatedAsyncioTestCase):
    @patch("soil.aio.dictionary.get_dictionary")
    async def test_get_dictionary(self, mock_get_dictionary):
        mock_get_dictionary.return_value = {"key": "value"}
        result = await dictionary("test_dict", "en")
        mock_get_dictionary.assert_called_once_with("test_dict", "en")
        self.assertEqual(result, {"key": "value"})

    @patch("soil.aio.dictionary.create_dictionary")
    async def test_create_dictionary(self, mock_create_dictionary):
        mock_create_dictionary.return_value = {"status": "created"}
        content = {"word1": "definition1", "word2": "definition2"}
        result = await dictionary("new_dict", "fr", content)
        mock_create_dictionary.assert_called_once_with("new_dict", "fr", content)
        self.assertEqual(result, {"status": "created"})

    async def test_invalid_input(self):
        with self.assertRaises(AssertionError):
            await dictionary(
                123,  # pyright:ignore[reportArgumentType]
                "en",
            )  # Invalid name type

        with self.assertRaises(AssertionError):
            await dictionary(
                "test_dict",
                123,  # pyright:ignore[reportArgumentType] Invalid language type
            )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring,line-too-long, unused-argument
# type: ignore

import unittest
from unittest.mock import MagicMock

from soil import task, task_wait


class TestTask(unittest.TestCase):
    def test_task(self):
        mock_module = MagicMock()
        mock_module.return_value = "ok"
        tasked = task(mock_module)
        future = tasked(1, 2, 3, a=2, b=3)
        result = task_wait(future)
        assert result == "ok"
        mock_module.assert_called_once_with(1, 2, 3, a=2, b=3)



"""Asynchronous Job module"""

from datetime import UTC, datetime

from soil.aio.api import create_experiment
from soil.data_structure import DataStructure
from soil.data_structure import DataStructure as InternalDataStructure
from soil.job import _Job
from soil.types import Job


async def _new_job(
    data_structure: InternalDataStructure,
    group: str,
    description: dict | str | None = None,
) -> Job:
    if data_structure.pipeline is None:
        raise ValueError("Pipeline plan is required")
    if data_structure.sym_id is None:
        raise ValueError("sym_id is required")
    experiment = await create_experiment(
        data_structure.pipeline.plan,
        experiment_group=group,
        description=description,
    )
    return _Job(
        experiment_id=experiment["_id"],
        result_id=experiment["outputs"][data_structure.sym_id],
        _created_at=datetime.fromtimestamp(
            experiment["created_at"] / 1000, tz=UTC
        ).isoformat(),
        _group=group,
    )


async def job(
    data_object: DataStructure, *, group: str, description: dict | str | None = None
) -> Job:
    """Creates a non-blocking job at soil."""
    return await _new_job(data_object, group=group, description=description)



"""Job module"""

from dataclasses import dataclass
from datetime import UTC, datetime

from soil import data
from soil.aio import api as aio_api
from soil.aio import data as aio_data
from soil.api import create_experiment, get_experiment
from soil.data_structure import DataStructure
from soil.data_structure import DataStructure as InternalDataStructure
from soil.types import ExperimentStatuses as Statuses
from soil.types import Job


@dataclass(kw_only=True, slots=True)
class _Job(Job):
    """Represents a Job."""

    experiment_id: str
    result_id: str
    _created_at: str
    _group: str

    @property
    def group(self) -> str:
        """Get the group of the job"""
        return self._group

    @property
    def created_at(self) -> str:
        """Get the created at of the job"""
        return self._created_at

    @property
    def status(self) -> Statuses:
        """Get the status of the job"""
        experiment = get_experiment(self.experiment_id)
        output_key = next(
            output_key
            for (output_key, output_id) in experiment["outputs"].items()
            if output_id == self.result_id
        )
        return experiment["status"][output_key]

    @property
    def data(self) -> DataStructure:  # type:ignore[reportIncompatibleMethodOverride]
        """Get the data of the job"""
        return data(self.result_id)

    async def async_data(self) -> DataStructure:
        """Get the data of the job asynchronously"""
        return await aio_data(self.result_id)

    async def async_status(self) -> Statuses:
        """Get the status of the job asynchronously"""
        experiment = await aio_api.get_experiment(self.experiment_id)
        output_key = next(
            output_key
            for (output_key, output_id) in experiment["outputs"].items()
            if output_id == self.result_id
        )
        return experiment["status"][output_key]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _Job):
            return False
        return (
            self.experiment_id == other.experiment_id
            and self.result_id == other.result_id
            and self._created_at == other._created_at
            and self._group == other._group
        )


def _new_job(
    data_structure: InternalDataStructure,
    group: str,
    description: dict | str | None = None,
) -> Job:
    if data_structure.pipeline is None:
        raise ValueError("Pipeline plan is required")
    if data_structure.sym_id is None:
        raise ValueError("sym_id is required")
    experiment = create_experiment(
        data_structure.pipeline.plan,
        experiment_group=group,
        description=description,
    )
    return _Job(
        experiment_id=experiment["_id"],
        result_id=experiment["outputs"][data_structure.sym_id],
        _created_at=datetime.fromtimestamp(
            experiment["created_at"] / 1000, tz=UTC
        ).isoformat(),
        _group=group,
    )


def job(
    data_object: DataStructure, *, group: str, description: dict | str | None = None
) -> Job:
    """Creates a non-blocking job at soil."""
    return _new_job(data_object, group=group, description=description)



"""Types used by soil library"""

# pylint:disable=unnecessary-ellipsis
from enum import StrEnum
from typing import (
    Dict,
    List,
    Literal,
    NotRequired,
    Protocol,
    Self,
    Type,
    TypedDict,
    TypeVar,
)

from numpy import ndarray
from pandas import DataFrame

from soil.storage.base_storage import BaseStorage

DEFAULT_TIMEOUT = 60 * 60

Plan = List[Dict[str, str]]

_TypedDict = TypeVar("_TypedDict", bound=TypedDict)  # type: ignore
type DataObject = dict | list | ndarray | DataFrame | _TypedDict


class GetModule(TypedDict):
    """Type for GET modules/:moduleId"""

    is_package: bool
    public_api: List[str]
    package_type: str


class GetModuleHash(TypedDict):
    """Type for GET modules/"""

    name: str
    hash: str


type ExperimentStatuses = Literal["WAITING", "EXECUTING", "DONE", "ERROR"]


class Experiment(TypedDict):
    """Type for GET experiments/:experimentId"""

    _id: str
    app_id: str
    outputs: Dict[str, str]
    experiment_status: ExperimentStatuses
    experiment_group: NotRequired[str]
    status: Dict[str, ExperimentStatuses]
    created_at: int


class Result(TypedDict):
    """Type for GET results/:resultId"""

    _id: str
    type: str


class _TypedDict(TypedDict):
    """Makes pyright happy"""

    pass


class SerializableDataStructure[Storage: BaseStorage, MetadataDict: _TypedDict](
    Protocol
):
    """Data Strucutre base protocol"""

    metadata: MetadataDict

    def serialize(self) -> Storage:
        """Serializes the DS."""
        ...

    @classmethod
    def deserialize(
        cls: Type[Self],
        storage: Storage,
        metadata: MetadataDict,
    ) -> "SerializableDataStructure[Storage, MetadataDict]":
        """Deserialize DS method."""
        ...


class DataStructure(Protocol):
    """Data Structure Protocol"""

    @property
    def metadata(self) -> dict:
        """Get the metadata of the DS"""
        ...

    @property
    def data(self) -> dict:
        """Get the data of the DS"""
        ...

    def get_data(self, **kwargs) -> dict:
        """Get the data of the DS"""
        ...


type JobStatuses = ExperimentStatuses


class Job(Protocol):
    """Job Protocol"""

    @property
    def group(self) -> str:
        """Get the group of the job"""
        ...

    @property
    def created_at(self) -> str:
        """Get the created at of the job"""
        ...

    @property
    def status(self) -> JobStatuses:
        """Get the status of the job"""
        ...

    @property
    def data(self) -> DataStructure:
        """Get the data of the job"""
        ...


class TypeLog(StrEnum):
    """Types of Logs."""

    PROCESSED = "processed"
    NOT_CONSISTENT = "not_consistent"
    NOT_PROCESSED = "not_processed"



"""Package that contains calls for soil get dictionary"""
from typing import Any, Dict, Optional

from soil.api import create_dictionary, get_dictionary


def dictionary(
    name: str, language: str, content: Optional[Dict[str, str]] = None
) -> Dict[str, Any]:
    """Create or get a dictionary"""
    if content is None:
        assert isinstance(name, str)
        assert isinstance(language, str)
        return get_dictionary(name, language)
    return create_dictionary(name, language, content)



"""Module generates"""

import asyncio
import os.path
from hashlib import sha256
from itertools import islice
from pathlib import Path
from typing import Generator

from soil.aio import api
from soil.types import GetModuleHash

FOLDER_BLACKLIST = [".vscode", ".venv", "test", ".git"]


BLOCK_SIZE = 1048576
MODULES_CHUNK = 25


def _hex_digest(data: bytes):
    file_hash = sha256()
    cursor = 0
    while cursor < len(data):
        file_hash.update(data[cursor : cursor + BLOCK_SIZE])
        cursor += BLOCK_SIZE
    return file_hash.hexdigest()


def _get_included_files() -> Generator[Path, None, None]:
    folders = [
        path
        for path in Path(".").glob("*")
        if path.is_dir() and path.name not in FOLDER_BLACKLIST
    ]
    for folder in folders:
        for selected_file in folder.rglob("*.py"):
            yield selected_file


def _filter_files(
    included_files: Generator[Path, None, None], file_hashes: list[GetModuleHash]
) -> Generator[tuple[str, str, bool], None, None]:
    file_hashes_dict = {module["name"]: module["hash"] for module in file_hashes}
    for file in included_files:
        code = file.read_text()
        code_hash = _hex_digest(bytes(code, encoding="utf-8"))
        module_name = str(file)[:-3].replace("/", ".")  # all are python files
        if file.stem == "__init__":
            module_name = module_name[: -len(".__init__")]
        if file_hashes_dict.get(module_name) != code_hash:
            yield (module_name, code, file.stem == "__init__")


async def _upload_selected_modules(
    modules: Generator[tuple[str, str, bool], None, None]
) -> None:
    mods = list(islice(modules, MODULES_CHUNK))
    tasks = []
    while len(mods) > 0:
        umods = []
        for module in mods:
            name, code, is_package = module
            umods.append({"name": name, "code": code, "is_package": is_package})
        tasks.append(api.upload_modules(modules=umods))
        mods = list(islice(modules, MODULES_CHUNK))
    await asyncio.gather(*tasks)


async def _upload_modules() -> None:
    modules = await api.get_modules()
    files_to_check = _get_included_files()
    modules_to_upload = _filter_files(files_to_check, modules)
    await _upload_selected_modules(modules_to_upload)
    await api.clean_client()


def upload_modules() -> None:
    """Upload modules to soil when not in a test environment."""
    if os.environ.get("PY_ENV", "development") == "test":
        return
    loop = None
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        asyncio.run(_upload_modules())
    if loop is not None:
        loop.create_task(_upload_modules())  # noqa: RUF006



# ''' Package for Elastic Search Connector '''  # noqa: D100
# This package is deprecated and can be removed.
# from typing import Iterable
# from typing import Any, Optional


# class ElasticSearchDBObject:
#     ''' Represents an Elastic Search Data Base Object. '''
#     # pylint: disable=unused-argument,redefined-builtin
#     def bulk(self, actions: Iterable[Any]) -> Any:
#         ''' Performs actions in bulk as defined in the ES bulk API.'''

#     def query(self, query: Any, random_sorting: bool = False) -> Any:
#         ''' Launches a query to ES. Returns the query.'''

#     def insert(self, document: Any, id: Optional[str] = None) -> Any:
#         ''' Inserts a document to ES. '''

#     def update_query(self, new_partial_query: Any) -> Any:
#         ''' Updates a query without executing it.'''


# def create_db_object(index: str, force_rewrite: bool = False) -> ElasticSearchDBObject:
#     ''' Obtains an Elastic Search DB object. Only works on cloud.'''
#     # pylint: disable=unused-argument
#     return ElasticSearchDBObject()



# pylint: disable=missing-docstring,line-too-long,broad-exception-raised, unnecessary-ellipsis
import unittest
from dataclasses import dataclass
from json import dumps, loads
from typing import Dict, NamedTuple, Optional, Self, Type, TypedDict
from unittest.mock import MagicMock, call, patch
from urllib.parse import parse_qs, urlparse

import soil
from soil.data_structure import DataStructure
from soil.storage.base_storage import BaseStorage


class MockHttpResponse(NamedTuple):
    """Soil configuration class"""

    status_code: int
    text: str

    def json(self) -> dict:
        return loads(self.text)


def set_id(self: DataStructure, timeout: int) -> None:
    self.id = "mock_id"
    self.timeout = timeout  # type:ignore[reportAttributeAccessIssue]


# pylint: disable=unused-argument
def mock_http_patch(
    url: str,
    *,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[Dict[str, str]] = None,
    timeout: int,
) -> MockHttpResponse:
    assert json is not None
    return MockHttpResponse(status_code=200, text=dumps({}))


# pylint: disable=unused-argument
def mock_http_post(
    url: str,
    *,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[Dict[str, str]] = None,
    timeout: int,
) -> MockHttpResponse:
    assert json is not None
    url_parts = urlparse(url)
    if url_parts.path == "/v2/results/":
        response = {"_id": "mock_id", "type": json["type"]}
        return MockHttpResponse(status_code=201, text=dumps(response))
    if url_parts.path == "/v2/states/":
        return MockHttpResponse(status_code=200, text=dumps({}))
    raise Exception("mock http case not found")


# pylint: disable=unused-argument
def mock_http_get(
    url: str, *, headers: Optional[Dict[str, str]] = None, timeout: int
) -> MockHttpResponse:
    url_parts = urlparse(url)
    query_params = parse_qs(url_parts.query)
    if url_parts.path == "/v2/states/" and query_params["name"][0] == "test_alias":
        return MockHttpResponse(status_code=200, text=dumps([]))
    raise Exception("mock http case not found")


# pylint: disable=unused-argument
def mock_http_get2(
    url: str, *, headers: Optional[Dict[str, str]] = None, timeout: int
) -> MockHttpResponse:
    url_parts = urlparse(url)
    query_params = parse_qs(url_parts.query)
    if url_parts.path == "/v2/states/" and query_params["name"][0] == "valid_alias":
        return MockHttpResponse(status_code=200, text=dumps([{"_id": "alias_id"}]))
    if url_parts.path == "/v2/states/alias_id/":
        result = {
            "_id": "alias_id",
            "state": {"result_id": "non_existing_alias_but_valid_id"},
        }
        return MockHttpResponse(status_code=200, text=dumps(result))
    raise Exception("mock http case not found")


def mock_http_get3(
    url: str, *, headers: Optional[Dict[str, str]] = None, timeout: int
) -> MockHttpResponse:
    if url == "test_auth_url/api/application/test_auth_app_id":
        return MockHttpResponse(
            status_code=200,
            text=dumps(
                {"application": {"roles": [{"name": "test"}, {"name": "test1"}]}}
            ),
        )
    raise Exception("mock http case not found")


def mock_http_post2(
    url: str,
    *,
    headers: Optional[Dict[str, str]] = None,
    json: dict | None = None,
    timeout: int,
) -> MockHttpResponse:
    if url == "test_auth_url/api/application/test_auth_app_id/role":
        return MockHttpResponse(status_code=200, text="ack")
    raise Exception("mock http case not found")


class TestAlias(unittest.TestCase):
    @patch("soil.api.session.get", side_effect=mock_http_get)
    @patch("soil.api.session.post", side_effect=mock_http_post)
    @patch("soil.api.session.patch", side_effect=mock_http_patch)
    def test_alias_new(
        self, mock_patch: MagicMock, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data([1, 2, 3, 4])
        soil.alias("test_alias", data)
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=test_alias&with-roles",
                    timeout=30,
                )
            ]
        )
        mock_post.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/results/",
                    json={
                        "type": "soil.data_structures.predefined.list.List",
                        "data": "[1, 2, 3, 4]",
                        "metadata": None,
                    },
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/states/",
                    json={
                        "name": "test_alias",
                        "state": {"alias": "test_alias", "result_id": "mock_id"},
                    },
                    timeout=30,
                ),
            ]
        )
        assert mock_patch.call_count == 0

    @patch("soil.api.session.get", side_effect=mock_http_get)
    @patch("soil.api.session.post", side_effect=mock_http_post)
    @patch("soil.api.session.patch", side_effect=mock_http_patch)
    def test_alias_new_with_type_hint(
        self, mock_patch: MagicMock, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        @dataclass(kw_only=True)
        class MockClass:
            class Metadata(TypedDict):
                test: str

            metadata: "MockClass.Metadata"
            storage: BaseStorage

            def serialize(self) -> BaseStorage:
                """Serializes the DS."""
                ...

            @classmethod
            def deserialize(
                cls: Type[Self],
                storage: BaseStorage,
                metadata: "MockClass.Metadata",
            ) -> Self:
                """Deserialize DS method."""
                ...

        data = soil.data([1, 2, 3, 4], return_type=MockClass)
        soil.alias("test_alias", data)
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=test_alias&with-roles",
                    timeout=30,
                )
            ]
        )
        mock_post.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/results/",
                    json={
                        "type": "soil.data_structures.predefined.list.List",
                        "data": "[1, 2, 3, 4]",
                        "metadata": None,
                    },
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/states/",
                    json={
                        "name": "test_alias",
                        "state": {"alias": "test_alias", "result_id": "mock_id"},
                    },
                    timeout=30,
                ),
            ]
        )
        assert mock_patch.call_count == 0

    @patch("soil.api.session.get", side_effect=mock_http_get2)
    @patch("soil.api.session.post", side_effect=mock_http_post)
    @patch("soil.api.session.patch", side_effect=mock_http_patch)
    def test_alias_existing(
        self, mock_patch: MagicMock, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data([1, 2, 3, 4])
        soil.alias("valid_alias", data)
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=valid_alias&with-roles",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    timeout=30,
                ),
            ]
        )
        mock_post.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/results/",
                    json={
                        "type": "soil.data_structures.predefined.list.List",
                        "data": "[1, 2, 3, 4]",
                        "metadata": None,
                    },
                    timeout=30,
                )
            ]
        )
        mock_patch.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    json={
                        "name": "valid_alias",
                        "state": {"alias": "valid_alias", "result_id": "mock_id"},
                    },
                    timeout=30,
                )
            ]
        )

    @patch("soil.api.session.get", side_effect=mock_http_get2)
    @patch("soil.api.session.post", side_effect=mock_http_post)
    @patch("soil.api.session.patch", side_effect=mock_http_patch)
    def test_alias_data_structure(
        self, mock_patch: MagicMock, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        with patch.object(
            DataStructure, "get_id", side_effect=set_id, autospec=True
        ) as mock_get_data:
            data = DataStructure(dstype="data_structures.some.test.type")
            soil.alias("valid_alias", data)
            assert mock_get_data.call_count == 1
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=valid_alias&with-roles",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    timeout=30,
                ),
            ]
        )
        assert mock_post.call_count == 0
        mock_patch.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    json={
                        "name": "valid_alias",
                        "state": {"alias": "valid_alias", "result_id": "mock_id"},
                    },
                    timeout=30,
                )
            ]
        )
        self.assertEqual(mock_get_data.call_args_list[0][1]["timeout"], 3600)

    @patch("soil.api.session.get", side_effect=mock_http_get2)
    @patch("soil.api.session.post", side_effect=mock_http_post)
    @patch("soil.api.session.patch", side_effect=mock_http_patch)
    def test_alias_data_structure_timeout(
        self, mock_patch: MagicMock, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        with patch.object(
            DataStructure, "get_id", side_effect=set_id, autospec=True
        ) as mock_get_data:
            data = DataStructure(dstype="data_structures.some.test.type")
            soil.alias("valid_alias", data, timeout=15)
            assert mock_get_data.call_count == 1
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=valid_alias&with-roles",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    timeout=30,
                ),
            ]
        )
        assert mock_post.call_count == 0
        mock_patch.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    json={
                        "name": "valid_alias",
                        "state": {"alias": "valid_alias", "result_id": "mock_id"},
                    },
                    timeout=30,
                )
            ]
        )
        self.assertEqual(mock_get_data.call_args_list[0][1]["timeout"], 15)

    @patch("soil.api.requests.post", side_effect=mock_http_post2)
    @patch("soil.api.requests.get", side_effect=mock_http_get3)
    @patch("soil.api.session.get", side_effect=mock_http_get2)
    @patch("soil.api.session.post", side_effect=mock_http_post)
    @patch("soil.api.session.patch", side_effect=mock_http_patch)
    def test_alias_roles(
        self,
        mock_patch: MagicMock,
        mock_post: MagicMock,
        mock_get: MagicMock,
        mock_requests_get: MagicMock,
        mock_requests_post: MagicMock,
    ) -> None:
        with patch.object(
            DataStructure, "get_id", side_effect=set_id, autospec=True
        ) as mock_get_data:
            data = DataStructure(dstype="data_structures.some.test.type")
            soil.alias("valid_alias", data, roles=["test"])
            assert mock_get_data.call_count == 1
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=valid_alias&with-roles",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    timeout=30,
                ),
            ]
        )
        assert mock_post.call_count == 0
        mock_patch.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    json={
                        "name": "valid_alias",
                        "state": {"alias": "valid_alias", "result_id": "mock_id"},
                        "roles": ["test"],
                    },
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/results/mock_id/",
                    json={"roles": ["test"]},
                    timeout=30,
                ),
            ]
        )
        self.assertListEqual(
            mock_requests_get.call_args_list,
            [
                call(
                    "test_auth_url/api/application/test_auth_app_id",
                    headers={
                        "Content-Type": "application/json",
                        "Authorization": "TEST_APP_INFO_API_KEY",
                    },
                    timeout=15,
                )
            ],
        )
        self.assertListEqual(mock_requests_post.call_args_list, [])

    @patch("soil.api.requests.post", side_effect=mock_http_post2)
    @patch("soil.api.requests.get", side_effect=mock_http_get3)
    @patch("soil.api.session.get", side_effect=mock_http_get)
    @patch("soil.api.session.post", side_effect=mock_http_post)
    @patch("soil.api.session.patch", side_effect=mock_http_patch)
    def test_alias_new_with_roles(
        self,
        mock_patch: MagicMock,
        mock_post: MagicMock,
        mock_get: MagicMock,
        mock_requests_get: MagicMock,
        mock_requests_post: MagicMock,
    ) -> None:
        data = soil.data([1, 2, 3, 4])
        soil.alias("test_alias", data, roles=["test1", "test2"])
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=test_alias&with-roles",
                    timeout=30,
                )
            ]
        )
        mock_post.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/results/",
                    json={
                        "type": "soil.data_structures.predefined.list.List",
                        "data": "[1, 2, 3, 4]",
                        "metadata": None,
                    },
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/states/",
                    json={
                        "name": "test_alias",
                        "state": {"alias": "test_alias", "result_id": "mock_id"},
                        "roles": ["test1", "test2"],
                    },
                    timeout=30,
                ),
            ]
        )
        mock_patch.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/results/mock_id/",
                    json={"roles": ["test1", "test2"]},
                    timeout=30,
                )
            ]
        )
        self.assertListEqual(
            mock_requests_get.call_args_list,
            [
                call(
                    "test_auth_url/api/application/test_auth_app_id",
                    headers={
                        "Content-Type": "application/json",
                        "Authorization": "TEST_APP_INFO_API_KEY",
                    },
                    timeout=15,
                ),
                call(
                    "test_auth_url/api/application/test_auth_app_id",
                    headers={
                        "Content-Type": "application/json",
                        "Authorization": "TEST_APP_INFO_API_KEY",
                    },
                    timeout=15,
                ),
            ],
        )
        self.assertListEqual(
            mock_requests_post.call_args_list,
            [
                call(
                    "test_auth_url/api/application/test_auth_app_id/role",
                    headers={
                        "Content-Type": "application/json",
                        "Authorization": "TEST_APP_INFO_API_KEY",
                    },
                    json={"role": {"name": "test2"}},
                    timeout=15,
                )
            ],
        )

    @patch("soil.api.requests.post", side_effect=mock_http_post2)
    @patch("soil.api.requests.get", side_effect=mock_http_get3)
    @patch("soil.api.session.get", side_effect=mock_http_get2)
    @patch("soil.api.session.post", side_effect=mock_http_post)
    @patch("soil.api.session.patch", side_effect=mock_http_patch)
    def test_alias_roles_and_extras(
        self,
        mock_patch: MagicMock,
        mock_post: MagicMock,
        mock_get: MagicMock,
        mock_requests_get: MagicMock,
        mock_requests_post: MagicMock,
    ) -> None:
        with patch.object(
            DataStructure, "get_id", side_effect=set_id, autospec=True
        ) as mock_get_data:
            data = DataStructure(dstype="data_structures.some.test.type")
            soil.alias("valid_alias", data, roles=["test"], extras={"name": "My Data"})
            assert mock_get_data.call_count == 1
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=valid_alias&with-roles",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    timeout=30,
                ),
            ]
        )
        assert mock_post.call_count == 0
        mock_patch.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    json={
                        "name": "valid_alias",
                        "state": {
                            "alias": "valid_alias",
                            "result_id": "mock_id",
                            "extras": {"name": "My Data"},
                        },
                        "roles": ["test"],
                    },
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/results/mock_id/",
                    json={"roles": ["test"]},
                    timeout=30,
                ),
            ]
        )
        self.assertListEqual(
            mock_requests_get.call_args_list,
            [
                call(
                    "test_auth_url/api/application/test_auth_app_id",
                    headers={
                        "Content-Type": "application/json",
                        "Authorization": "TEST_APP_INFO_API_KEY",
                    },
                    timeout=15,
                )
            ],
        )
        self.assertListEqual(mock_requests_post.call_args_list, [])



"""Module for the compound storage"""

from dataclasses import dataclass
from typing import List, Tuple

from soil.storage.base_storage import BaseStorage


@dataclass
class CompoundStorage[K: str, V: BaseStorage](BaseStorage):
    """A meta storage composed of other storages as a dict.

    Example:
    -------
        compound = CompoundStorage(storages={
            "db": Elasticsearch(index="index1"),
            "disk": ObjectStorage()
        })
        compound["second_db"] = ElasticSearch(index="index2")
        compound["db"].search(query=myquery)
    """

    storages: dict[K, V]

    def __getitem__(self, storage_name: str) -> BaseStorage:
        """Return the storage with that storage_name"""

    def __setitem__(self, storage_name: str, storage: BaseStorage) -> None:
        """Set storage to that storage_name"""

    def __len__(self) -> int:
        """Returns the number of storages"""

    def items(self) -> List[Tuple[str, BaseStorage]]:
        """Returns an iterable of sotrage_name, storage tuples"""



"""Module that defines a Pipeline."""

# To prevent Pipeline not defined: https://stackoverflow.com/a/49872353/3481480
from __future__ import annotations

import asyncio
import copy
import datetime
import logging
from time import sleep, time
from typing import Any, Dict, List, Optional

from soil import api, errors
from soil.aio import api as aio_api
from soil.logger import logger as soil_logger
from soil.types import DEFAULT_TIMEOUT, Experiment, Plan

# How much should wait between api calls
# Remember ES takes some time to index logs
SLEEP_TIME = 1
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Pipeline:
    """A Pipeline stores the transformations and dependencies to obtain certain results."""

    def __init__(self, plan: Optional[Plan] = None) -> None:  # noqa: D107
        self.plan = plan if plan is not None else []
        self.experiment: Optional[Experiment] = None
        self.finished = False

    def run(self, timeout: int = DEFAULT_TIMEOUT) -> Dict[str, str]:
        """Run the Pipeline (blocking call until the experiment finishes or timesout"""
        if self.finished and self.experiment:
            return self.experiment["outputs"]
        if self.experiment is None:
            experiment = api.create_experiment(self.plan)
            self.experiment = experiment
        status = api.get_experiment(experiment["_id"])["experiment_status"]
        start_date = datetime.datetime.now().astimezone().isoformat()
        max_time = time() + timeout
        while status not in ["DONE", "ERROR"]:
            if time() > max_time:
                raise errors.ExperimentTimeoutError("Pipeline timed out")
            sleep(SLEEP_TIME)
            logs = api.get_experiment_logs(experiment["_id"], start_date)
            _print_logs(logs)
            if len(logs) > 0:
                start_date = logs[0]["date"]
            status = api.get_experiment(experiment["_id"])["experiment_status"]
        sleep(SLEEP_TIME)
        logs = api.get_experiment_logs(experiment["_id"], start_date)
        _print_logs(logs)
        if status == "ERROR":
            raise errors.ExperimentError("Pipeline failed")
        logger.debug("experiment_done: %s", experiment["_id"])
        self.finished = True
        return experiment["outputs"]

    def add_transformation(self, transformation: Dict[str, str]) -> Pipeline:
        """Add a new transformation to the Pipeline, returns a new Pipeline
        containing the plan of the old Pipeline plus the transformation.
        """
        new_plan = [*self.plan, transformation]
        return Pipeline(plan=new_plan)

    @staticmethod
    def merge_pipelines(*pipelines: Pipeline) -> Pipeline:
        """Merges all the Pipelines passed into a new Pipeline that is returned."""
        merged_plan: Plan = sum([p.plan for p in pipelines], [])
        # Remove repeated
        merged_plan = [
            i for n, i in enumerate(merged_plan) if i not in merged_plan[n + 1 :]
        ]
        return Pipeline(plan=merged_plan)

    async def async_run(self, timeout: int = DEFAULT_TIMEOUT) -> Dict[str, str]:
        """Run the Pipeline asynchronously until the experiment finishes or times out."""
        if self.finished and self.experiment:
            return self.experiment["outputs"]
        if self.experiment is None:
            self.experiment = await aio_api.create_experiment(self.plan)

        status = (await aio_api.get_experiment(self.experiment["_id"]))[
            "experiment_status"
        ]
        start_date = datetime.datetime.now().astimezone().isoformat()
        max_time = time() + timeout

        while status not in ["DONE", "ERROR"]:
            if time() > max_time:
                raise errors.ExperimentTimeoutError("Pipeline timed out")
            await asyncio.sleep(SLEEP_TIME)
            logs = await aio_api.get_experiment_logs(self.experiment["_id"], start_date)
            _print_logs(logs)
            if logs:
                start_date = logs[0]["date"]
            status = (await aio_api.get_experiment(self.experiment["_id"]))[
                "experiment_status"
            ]

        await asyncio.sleep(SLEEP_TIME)
        logs = await aio_api.get_experiment_logs(self.experiment["_id"], start_date)
        _print_logs(logs)

        if status == "ERROR":
            raise errors.ExperimentError("Pipeline failed")
        logger.debug("experiment_done: %s", self.experiment["_id"])
        self.finished = True
        return self.experiment["outputs"]


def _print_logs(logs: List[Dict[str, Any]]) -> None:
    for log in logs[::-1]:
        level = getattr(logging, log["level"], logging.INFO)
        nlog = copy.copy(log)
        del nlog["message"]
        soil_logger.log(level, "%s - %s", nlog["date"], log["message"], extra=nlog)



"""Module that implements a Data Structure"""

import json
import os
from typing import Any, Dict, Optional, Tuple, Union, cast

from soil import api, errors
from soil.aio import api as aio_api
from soil.pipeline import Pipeline
from soil.types import DEFAULT_TIMEOUT, DataObject

try:
    import numpy  # type: ignore

    NUMPY_ENABLED = True
except ModuleNotFoundError:
    NUMPY_ENABLED = False
try:
    import pandas  # type: ignore

    PANDAS_ENABLED = True
except ModuleNotFoundError:
    PANDAS_ENABLED = False


DS_NAMESPACE = "soil.data_structures.predefined."


class DataStructure[D: DataObject]:
    """Data Structure class"""

    def __init__(  # noqa: D107
        self,
        did: Optional[str] = None,
        sym_id: Optional[str] = None,
        data: Optional[D] = None,
        metadata: Optional[Dict[str, Any]] = None,
        pipeline: Optional[Pipeline] = None,
        dstype: Optional[str] = None,
    ):
        self.id = did  # Result id
        self.sym_id = sym_id  # Symbolic id used in the pipeline
        self.data = cast(D, data)
        self.metadata = metadata
        self.pipeline = pipeline
        self.dstype = dstype

    async def async_get_data(
        self, **kwargs: Dict[str, Any]
    ) -> Dict[Union[str, int], Any]:
        """Asynchronously invoke the get_data() method from the data structure in the cloud."""
        result_id = await self.async_get_id()
        result = await aio_api.get_result_data(
            result_id, query={k: json.dumps(v) for k, v in kwargs.items()}
        )
        self.metadata = result["metadata"]
        self.dstype = result["type"]
        return _unserialize_data(result["type"], result["results"])

    async def async_get_id(self, timeout: int = DEFAULT_TIMEOUT) -> str:
        """Asynchronously get the result ID."""
        if self.id is None:
            if not self.pipeline:
                raise errors.DataStructurePipelineNotFound(
                    "Data structure has no pipeline to run."
                )
            results = await self.pipeline.async_run(timeout=timeout)
            assert isinstance(self.sym_id, str)
            self.id = results[self.sym_id]
        return self.id

    async def async_get_metadata(self) -> Dict[str, Any]:
        """Asynchronously get the metadata of the data structure."""
        if self.metadata:
            return self.metadata

        result_id = await self.async_get_id()
        result = await aio_api.get_result(result_id)
        self.dstype = result["type"]
        self.metadata = result["metadata"]
        return self.metadata

    def get_data(self, **kwargs: Dict[str, Any]) -> Dict[Union[str, int], Any]:
        """Invoke the get_data() method from the data structure in the cloud."""  # noqa: D402
        result_id = self.get_id()
        result = api.get_result_data(
            result_id, query={k: json.dumps(v) for k, v in kwargs.items()}
        )
        self.metadata = result["metadata"]
        self.dstype = result["type"]
        return _unserialize_data(result["type"], result["results"])

    def export(self, path: str, **kwargs: Dict[str, Any]) -> str:
        """Export a result to a file. The file will be stored in the folder
        and returns the file_path.
        """
        result_id = self.get_id()
        file_path = os.path.join(path, "result.zip")
        api.export_result(result_id, file_path, **kwargs)
        return file_path

    def get_id(self, timeout: int = DEFAULT_TIMEOUT) -> str:
        """Invoke the get_data() method from the data structure in the cloud."""
        result_id = self.id
        if result_id is None:
            if not self.pipeline:
                raise errors.DataStructurePipelineNotFound(
                    "Data structure has no pipeline to run."
                )
            results = self.pipeline.run(timeout=timeout)  # run
            assert isinstance(self.sym_id, str)
            result_id = results[self.sym_id]
            self.id = result_id
        return result_id

    def __getattribute__(self, name: str) -> Any:  # noqa: D105
        if name in ("data", "metadata"):
            if (object.__getattribute__(self, "data") and name == "data") or (
                object.__getattribute__(self, "metadata") and name == "metadata"
            ):
                return object.__getattribute__(self, name)
            result_id = self.id
            if not self.pipeline and result_id is None:
                raise errors.DataStructurePipelineNotFound(
                    "Data structure has no pipeline to run."
                )
            if result_id is None:
                assert isinstance(self.pipeline, Pipeline)
                assert isinstance(self.sym_id, str)
                results = self.pipeline.run()
                result_id = results[self.sym_id]
                self.id = result_id
            if name == "data":
                result = api.get_result_data(result_id)
                self.dstype = result["type"]
                self.data = _unserialize_data(result["type"], result["results"])
            elif name == "metadata":
                result = api.get_result(result_id)
                self.dstype = result["type"]
            self.dstype = result["type"]
            self.metadata = result["metadata"]
        return object.__getattribute__(self, name)


def _unserialize_data(dtype: str, data: Any) -> Any:
    if dtype == "data_structures.predefined.ndarray.Ndarray":
        return numpy.array(data)
    if dtype == "data_structures.predefined.pd_data_frame.PdDataFrame":
        return pandas.DataFrame(data=data)
    return data


def get_data_structure_name_and_serialize(data_object: Any) -> Tuple[str, str]:
    """Get the data structure and serialize it."""
    data_type = type(data_object)
    if data_type == dict:
        return DS_NAMESPACE + "dict.Dict", json.dumps(data_object)
    if data_type == list:
        return DS_NAMESPACE + "list.List", json.dumps(data_object)
    if NUMPY_ENABLED and data_type == numpy.ndarray:
        return DS_NAMESPACE + "ndarray.Ndarray", json.dumps(data_object.tolist())
    if PANDAS_ENABLED and data_type == pandas.core.frame.DataFrame:
        return DS_NAMESPACE + "pd_data_frame.PdDataFrame", data_object.to_json()
    raise errors.DataStructureType(f"Unrecognised type: {data_type}")
